#!/usr/bin/env python3
"""
The MIT License (MIT)
Copyright © 2015 RealDolos
See LICENSE
"""

import os
import sys
import uuid

from io import BytesIO
from collections.abc import Mapping
from urllib.parse import quote

try:
    import orjson as json
except ImportError:
    import json


def generate_boundary():
    """Generates a boundary string to be used for multipart/form-data"""

    return uuid.uuid4().hex


def escape_header(val):
    """Escapes a value so that it can be used in a mime header"""

    if val is None:
        return None
    try:
        return quote(val, encoding="ascii", safe="/ ")
    except ValueError:
        return "utf-8''" + quote(val, encoding="utf-8", safe="/ ")


def make_streams(name, value, boundary, encoding):
    """Generates one or more streams for each name, value pair"""

    filename = None
    mime = None

    # user passed in a special dict.
    if isinstance(value, Mapping) and "name" in value and "value" in value:
        filename = value["name"]
        try:
            mime = value["mime"]
        except KeyError:
            pass
        value = value["value"]

    if not filename:
        filename = getattr(value, "name", None)
        if filename:
            filename = os.path.split(filename)[1]

    mime = mime or "application/octet-stream"

    name, filename, mime = [escape_header(v) for v in (name, filename, mime)]

    stream = BytesIO()
    stream.write("--{}\r\n".format(boundary).encode(encoding))
    if not filename:
        stream.write(
            'Content-Disposition: form-data; name="{}"\r\n'.format(name).encode(
                encoding
            )
        )
    else:
        stream.write(
            'Content-Disposition: form-data; name="{}"; filename="{}"\r\n'.format(
                name, filename
            ).encode(encoding)
        )
        if mime:
            stream.write("Content-Type: {}\r\n".format(mime).encode(encoding))
    stream.write(b"\r\n")

    if hasattr(value, "read"):
        stream.seek(0)
        return stream, value, BytesIO("\r\n".encode(encoding))

    # not a file-like object, encode headers and value in one go
    value = value if isinstance(value, (str, bytes)) else json.dumps(value)
    if isinstance(value, bytes):
        stream.write(value)
    else:
        stream.write(value.encode(encoding))
    stream.write(b"\r\n")
    stream.seek(0)
    return (stream,)


class Data:
    """multipart/form-data generator

    The generator will own any file-like objects you pass in. You don't
    need to close such objects yourself.

    Supported value types:
     - POD types
     - file-like-objects
     - json-serializable objects (will be serialized as json

    You may specify file names and value types by passing special dicts
    that contain name, value and optionally mime keys. values will be
    treated like normal objects.

    Data objects must be iterated over (streamed). Each iteration result
    will contain at most blocksize bytes. This enables Data objects to
    encode multipart/form-data requests without having to read all data
    into memory at once
    """

    def __init__(
        self, values, blocksize=0, encoding="utf-8", callback=None, logical_offset=0
    ):
        self.encoding = encoding or "utf-8"
        self.boundary = generate_boundary()
        self.streams = []
        self.callback = callback or None
        self.blocksize = blocksize
        self.logical_offset = logical_offset
        if not self.blocksize or self.blocksize <= 0:
            self.blocksize = 1 << 17

        for name, value in values.items():
            self.streams.extend(make_streams(name, value, self.boundary, encoding))
        self.streams.append(
            BytesIO("--{}--\r\n".format(self.boundary).encode(encoding))
        )

    @property
    def len(self):
        """Length of the data stream"""
        # The len property is needed for requests.
        # requests checks __len__, then len
        # Since we cannot implement __len__ because python 32-bit uses 32-bit
        # sizes, we implement this instead.
        def stream_len(stream):
            """Stream length"""
            cur = stream.tell()
            try:
                stream.seek(0, 2)
                return stream.tell() - cur
            finally:
                stream.seek(cur)

        return sum(stream_len(s) for s in self.streams)

    @property
    def headers(self):
        """All headers needed to make a request"""
        return {
            "Content-Type": ("multipart/form-data; boundary={}".format(self.boundary)),
            "Content-Length": str(self.len),
            "Content-Encoding": self.encoding,
        }

    def __iter__(self):
        with self:
            total = None
            if self.callback:
                total = self.len
            pos = 0
            remainder = self.blocksize
            buf = BytesIO()
            stream = self.streams.pop(0)
            while stream:
                with stream:
                    while remainder:
                        cur = stream.read(remainder)
                        if not cur:
                            break
                        buf.write(cur)
                        remainder -= len(cur)
                        if not remainder:
                            val = buf.getvalue()
                            yield val
                            if self.callback:
                                pos += len(val)
                                self.callback(
                                    self.logical_offset + pos,
                                    self.logical_offset + total,
                                )
                            remainder = self.blocksize
                            buf = BytesIO()
                try:
                    stream = self.streams.pop(0)
                except IndexError:
                    break

            last = buf.getvalue()
            if not last:
                return

            pos += len(last)
            yield last
            if self.callback:
                self.callback(self.logical_offset + pos, self.logical_offset + total)

    def __enter__(self):
        return self

    def __exit__(self, _extype, _value, _traceback):
        self.close()

    def close(self):
        """Close multipart instance and all associated streams"""
        try:
            for stream in self.streams:
                with stream:
                    pass
        finally:
            del self.streams[:]


if __name__ == "__main__":
    import requests

    def my_callback(cur, tot):
        """multipart callback"""
        print(cur, tot, "{:.2%}".format(float(cur) / tot), file=sys.stderr)

    def main():
        """main()"""
        multipart = {
            "abc": "def",
            "123": 345,
            "flt": 1.0,
            "json": {"hellow": ["world"]},
            "special": {"name": "huhu русский язык 中文", "value": "value"},
            "file1": open(__file__, "rb"),
            "file2": {
                "name": "file.py",
                "value": open(__file__, "rb"),
                "mime": "application/x-python",
            },
        }
        multipart = Data(multipart, callback=my_callback, blocksize=100)

        if len(sys.argv) > 1:
            print(
                requests.post(
                    sys.argv[1], headers=multipart.headers, data=multipart
                ).content
            )
            sys.exit(0)

        for key, val in multipart.headers.items():
            print("{}: {}".format(key, val), file=sys.stderr)
        for i in multipart:
            sys.stdout.buffer.write(i)
            sys.stdout.flush()
        sys.exit(0)

    main()



import string


class User:
    """Used by Room. Currently not very useful by itself"""

    def __init__(self, nick, conn, max_len):
        self.__max_length = max_len
        if nick is None:
            self.nick = ""
        else:
            self.__verify_username(nick)
        self.nick = nick
        self.conn = conn
        self.logged_in = False
        self.session = None

    def login(self, password):
        """Attempts to log in as the current user with given password"""

        if self.logged_in:
            raise RuntimeError("User already logged in!")

        params = {"name": self.nick, "password": password}
        resp = self.conn.make_api_call("login", params)
        if "error" in resp:
            raise RuntimeError(
                f"Login failed: {resp['error'].get('message') or resp['error']}"
            )
        self.session = resp["session"]
        self.conn.make_call("useSession", self.session)
        self.conn.cookies.update({"session": self.session})
        self.logged_in = True
        return True

    def login_transplant(self, other):
        """Attempts to carry over the login state from another room"""

        if not other.logged_in:
            raise ValueError("Other room is not logged in")
        cookie = other.session
        if not cookie:
            raise ValueError("Other room has no cookie")
        self.conn.cookies.update({"session": cookie})
        self.session = cookie
        self.logged_in = True
        return True

    def logout(self):
        """Logs your user out"""

        if not self.logged_in:
            raise RuntimeError("User is not logged in")
        if self.conn.connected:
            params = {"room": self.conn.room.room_id}
            resp = self.conn.make_api_call("logout", params)
            if not resp.get("success", False):
                raise RuntimeError(
                    f"Logout unsuccessful: "
                    f"{resp['error'].get('message') or resp['error']}"
                )
            self.conn.make_call("logout", params)
            self.conn.cookies.pop("session")
        self.logged_in = False

    def change_nick(self, new_nick):
        """Change the name of your user
        Note: Must be logged out to change nick"""

        if self.logged_in:
            raise RuntimeError("User must be logged out")
        self.__verify_username(new_nick)

        self.conn.make_call("command", self.nick, "nick", new_nick)
        self.nick = new_nick

    def register(self, password):
        """Registers the current user with the given password."""

        if len(password) < 8:
            raise ValueError("Password must be at least 8 characters.")

        params = {"name": self.nick, "password": password}
        resp = self.conn.make_api_call("register", params)

        if "error" in resp:
            raise RuntimeError(f"{resp['error'].get('message') or resp['error']}")

        self.conn.make_call("useSession", resp["session"])
        self.conn.cookies.update({"session": resp["session"]})
        self.logged_in = True

    def __verify_username(self, username):
        """Raises an exception if the given username is not valid."""

        if len(username) > self.__max_length or len(username) < 3:
            raise ValueError(
                f"Username must be between 3 and {self.__max_length} characters."
            )
        if any(c not in string.ascii_letters + string.digits for c in username):
            raise ValueError("Usernames can only contain alphanumeric characters.")

    def __repr__(self):
        return f"<User({self.nick}, {self.logged_in})>"



class Config(dict):

    """Class that allows accessing room config dict by using attribute names.
    Mapping specifies what values will be read/written by our API."""

    def __init__(self):
        super().__init__()
        self.setdefault("janitors", list())
        self.setdefault("disabled", False)
        self.setdefault("owner", str())
        self.__cfg_mapping = dict(
            adult=("adult", bool),
            private=("private", bool),
            disabled=("disabled", bool),
            owner=("owner", str),
            janitors=("janitors", list),
            room_id=("room_id", str),
            alias=("custom_room_id", str),
            title=("name", str),
            motd=("motd", str),
            max_title=("max_room_name_length", int),
            max_message=("chat_max_message_length", int),
            max_nick=("chat_max_alias_length", int),
            max_file=("file_max_size", int),
            session_lifetime=("session_lifetime", int),
            file_ttl=("file_ttl", int),
            creation_time=("created_time", int),
        )

    def __getattr__(self, key):
        if key not in self.__cfg_mapping:
            raise KeyError(f"Mapping for the key `{key}` doesn't exist")
        return self[key]

    def update(self, config):
        for k, v in self.__cfg_mapping.items():
            if v[0] not in config:
                continue
            if isinstance(config[v[0]], v[1]):
                self[k] = config[v[0]]
            else:
                self[k] = v[1]()

    def get_real_key(self, key):
        return self.__cfg_mapping[key][0]



import time

from .constants import BASE_URL


class File:
    """Basically a struct for a file's info on volafile, with an additional
    method to retrieve the file's URL."""

    def __init__(self, room, conn, file_id, name, **kw):
        self.room = room
        self.conn = conn
        self.fid = file_id
        self.name = name
        # Flag to check if we queried `fileinfo` for this file already
        self.updated = False
        self.__additional = dict(kw)

    def __getattr__(self, name):
        if name not in (
            "filetype",
            "size",
            "expire_time",
            "uploader",
            "checksum",
            "info",
            "thumb",
        ):
            raise AttributeError(f"Not a valid key: {name}")
        try:
            return self.__additional[name]
        except KeyError:
            self.conn.queues_enabled = False
            self.room.fileinfo(self.fid)
            self.conn.queues_enabled = True
            return self.__additional[name]

    def fileupdate(self, data):
        """Method to update extra metadata fields with dict obtained
        through `fileinfo`"""

        self.name = data["name"]
        add = self.__additional
        add["filetype"] = "other"
        for filetype in ("book", "image", "video", "audio", "archive"):
            if filetype in data:
                add["filetype"] = filetype
                break
        if add["filetype"] in ("image", "video", "audio"):
            add["thumb"] = data.get("thumb", {})
        # checksum is md5
        add["checksum"] = data["checksum"]
        add["expire_time"] = data["expires"] / 1000
        add["size"] = data["size"]
        add["info"] = data.get(add["filetype"], {})
        add["uploader"] = data["user"]
        if self.room.admin:
            add["info"].update({"room": data.get("room")})
            add["info"].update({"uploader_ip": data.get("uploader_ip")})
        self.updated = True

    @property
    def url(self):
        """Gets the download url of the file"""

        return f"{BASE_URL}/get/{self.fid}/{self.name}"

    @property
    def expired(self):
        """Returns true if the file has expired, false otherwise"""

        return time.time() >= self.expire_time

    @property
    def time_left(self):
        """Returns how many seconds before this file expires"""

        return self.expire_time - time.time()

    @property
    def thumbnail(self):
        """Returns the thumbnail's url for this image, audio, or video file.
        Returns empty string if the file has no thumbnail"""

        if self.filetype not in ("video", "image", "audio"):
            raise RuntimeError("Only video, audio and image files can have thumbnails")
        thumb_srv = self.thumb.get("server")
        url = f"https://{thumb_srv}" if thumb_srv else None
        return f"{url}/asset/{self.fid}/thumb" if url else ""

    @property
    def resolution(self):
        """Gets the resolution of this image or video file in format (W, H)"""

        if self.filetype not in ("video", "image"):
            raise RuntimeError("Only videos and images have resolutions")
        return (self.info["width"], self.info["height"])

    @property
    def duration(self):
        """Returns the duration in seconds of this audio or video file"""

        if self.filetype not in ("video", "audio"):
            raise RuntimeError("Only videos and audio have durations")
        return self.info.get("length") or self.info.get("duration")

    @property
    def album(self):
        """Returns album name of audio file"""

        if self.filetype not in ("audio",):
            raise RuntimeError("Only audio files can have album names")
        return self.info.get("album")

    @property
    def artist(self):
        """Returns artist name of audio file"""

        if self.filetype not in ("audio",):
            raise RuntimeError("Only audio files can have artist names")
        return self.info.get("artist")

    @property
    def codec(self):
        """Returns codec type of media file"""

        if self.filetype not in ("video", "audio"):
            raise RuntimeError("Only audio and video files can have codecs")
        return self.info.get("codec")

    @property
    def title(self):
        """Returns title of media file"""

        if self.filetype not in ("video", "audio"):
            raise RuntimeError("Only audio and video files can have titles")
        return self.info.get("title")

    @property
    def ip_address(self):
        """Returns the uploader ip if you are a mod"""
        self.room.check_admin()
        return self.info.get("uploader_ip")

    @property
    def host_room(self):
        """Returns ID of the room to which the file is uploaded if you are a mod"""
        self.room.check_admin()
        return self.info.get("room")

    def delete(self):
        """ Remove this file """
        self.room.check_owner()
        self.conn.make_call("deleteFiles", [self.fid])

    def timeout(self, duration=3600):
        """ Timeout the uploader of this file """
        self.room.check_owner()
        self.conn.make_call("timeoutFile", self.fid, duration)

    def __repr__(self):
        return f"<File({self.fid}, {self.size}, {self.uploader}, {self.name})>"



#!/usr/bin/env python3

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

from volapi import Room


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print("Provide a room ID", file=sys.stderr)
        sys.exit(1)

    def callback(current, total):
        print("Uploaded {} of {} bytes or {:.2%}".
              format(current, total, float(current) / total))

    with Room(sys.argv[1]) as r:
        r.user.change_nick("DumbUpload")
        r.upload_file(__file__, callback=callback)
    print("done")




"""
This file is part of Volapi.

Volapi is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Volapi is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Volapi.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import warnings
import asyncio
import os
import re
import time

from collections import OrderedDict
from collections import defaultdict
from contextlib import suppress
from threading import get_ident as get_thread_ident
from threading import Barrier, RLock
from concurrent.futures import TimeoutError as FutureTimeoutError


import requests
from .auxo import ARBITRATOR, Listeners, Protocol
from .handler import Handler
from .config import Config
from .user import User
from .multipart import Data
from .utils import delayed_close, random_id, to_json, from_json
from .constants import __version__, MAX_UNACKED, BASE_URL, BASE_REST_URL, BASE_WS_URL

LOGGER = logging.getLogger(__name__)


class Connection(requests.Session):
    """Bundles a requests/websocket pair"""

    def __init__(self, room):

        super().__init__()

        self.room = room
        self.exception = None

        self.__lastping = self.__lastpong = 0

        agent = f"Volafile-API/{__version__}"

        self.headers.update({"User-Agent": agent})
        self.cookies.update({"allow-download": "1"})

        self.lock = RLock()
        self.__conn_barrier = Barrier(2)
        self.listeners = defaultdict(Listeners)
        self.must_process = False
        self.__queues_enabled = True

        self.__ping_interval = 20  # default

        self.proto = Protocol(self)
        self.handler = Handler(self)
        self.last_ack = self.proto.max_id

    def connect(self, username, checksum, password=None, key=None):
        """Connect to websocket through asyncio http interface"""

        ws_url = (
            f"{BASE_WS_URL}?room={self.room.room_id}&cs={checksum}&nick={username}"
            f"&rn={random_id(6)}&t={int(time.time() * 1000)}&transport=websocket&EIO=3"
        )
        if password:
            ws_url += f"&password={password}"
        elif key:
            ws_url += f"&key={key}"

        ARBITRATOR.create_connection(
            self.proto, ws_url, self.headers["User-Agent"], self.cookies
        )
        self.__conn_barrier.wait()

    @property
    def ping_interval(self):
        """Gets the ping interval"""

        return self.__ping_interval

    @ping_interval.setter
    def ping_interval(self, interval):
        self.__ping_interval = interval

    @property
    def connected(self):
        """Connection state"""

        return bool(hasattr(self, "proto") and self.proto.connected)

    def send_message(self, payload):
        """Send a message"""

        ARBITRATOR.send_async_message(self.proto, payload)

    def send_ack(self):
        """Send an ack message"""

        if self.last_ack == self.proto.max_id:
            return
        LOGGER.debug("ack (%d)", self.proto.max_id)
        self.last_ack = self.proto.max_id
        self.send_message(b'4' + to_json([self.proto.max_id]))

    def make_call(self, fun, *args):
        """Makes a regular API call"""

        obj = {"fn": fun, "args": list(args)}
        obj = [self.proto.max_id, [[0, ["call", obj]], self.proto.send_count]]
        self.send_message(b'4' + to_json(obj))
        self.proto.send_count += 1

    def make_call_with_cb(self, fun, *args):
        """Makes an API call with a callback to wait for"""

        async def call_and_wait():
            cid, event = self.handler.register_callback()
            argscp = list(args)
            argscp.append(cid)
            self.make_call(fun, *argscp)
            return await event
        return asyncio.run_coroutine_threadsafe(call_and_wait(), ARBITRATOR.loop)

    def make_api_call(self, call, params):
        """Make a REST API call"""

        if not isinstance(params, dict):
            raise ValueError("params argument must be a dictionary")
        kw = dict(
            params=params,
            headers={"Origin": BASE_URL, "Referer": f"{BASE_URL}/r/{self.room.name}"},
        )
        return self.get(BASE_REST_URL + call, **kw).json()

    def reraise(self, ex):
        """Reraise an exception passed by the event thread"""

        self.exception = ex
        self.process_queues(forced=True)

    def close(self):
        """Closes connection pair"""

        if self.connected:
            obj = [self.proto.max_id, [[2], self.proto.send_count]]
            ARBITRATOR.send_sync_message(self.proto, b'4' + to_json(obj))
            self.proto.send_count += 1
            ARBITRATOR.close(self.proto)
        self.listeners.clear()
        self.proto.connected = False
        super().close()
        del self.room
        del self.proto

    def __ensure_barrier(self):
        if self.__conn_barrier:
            self.__conn_barrier.wait()
            self.__conn_barrier = None

    async def on_open(self):
        """DingDongmaster the connection is open"""

        self.__ensure_barrier()
        while self.connected:
            try:
                if self.__lastping > self.__lastpong:
                    raise IOError("Last ping remained unanswered")

                self.send_message("2")
                self.send_ack()
                self.__lastping = time.time()
                await asyncio.sleep(self.ping_interval)
            except Exception as ex:
                LOGGER.exception("Failed to ping")
                try:
                    self.reraise(ex)
                except Exception:
                    LOGGER.exception(
                        "failed to force close connection after ping error"
                    )
                break

    async def on_close(self):
        """DingDongmaster the connection is gone"""

        self.__ensure_barrier()
        return None

    def _on_frame(self, data):
        if not hasattr(self, "room"):
            LOGGER.debug("received out of bounds message [%r]", data)
            return
        LOGGER.debug("received message %r", data)
        if not isinstance(data, list):
            self.proto.session = data["session"]
        elif isinstance(data, list) and len(data) > 1:
            data = data[1:]
            last_ack = int(data[-1][-1])
            self.proto.max_id = last_ack
            if last_ack > self.proto.max_id + MAX_UNACKED:
                LOGGER.debug("needing to ack (%d/%d)", last_ack, self.proto.max_id)
                self.send_ack()
            self.handler.add_data(data)
        elif data == [2]:
            LOGGER.debug("Server send close message")
            self.room.close()
        elif data == [0]:
            LOGGER.warning("Some IO Error, maybe reconnect after it?")
            # raise IOError("Force disconnect?")
        elif isinstance(data, list):
            # ignore
            pass
        # not really handled further
        else:
            LOGGER.warning("unhandled message frame type %r", data)

    def on_message(self, new_data):
        """Processes incoming messages according to engine-io rules"""
        # https://github.com/socketio/engine.io-protocol

        LOGGER.debug("new frame [%r]", new_data)
        try:
            what = int(new_data[0])
            data = new_data[1:]
            data = data and from_json(data)
            if what == 0:
                self.ping_interval = float(data["pingInterval"]) / 1000
                LOGGER.debug("adjusted ping interval")
                return

            if what == 1:
                LOGGER.debug("received close")
                self.reraise(IOError("Connection closed remotely"))
                return

            if what == 3:
                self.__lastpong = time.time()
                LOGGER.debug("received a pong")
                return

            if what == 4:
                self._on_frame(data)
                return

            if what == 6:
                LOGGER.debug("received noop")
                self.send_message("5")
                return

            LOGGER.debug("unhandled message: [%d] [%r]", what, data)
        except Exception as ex:
            self.reraise(ex)

    def add_listener(self, event_type, callback):
        """Add a listener for specific event type.
        You'll need to actually listen for changes using the listen method"""

        if not self.connected:
            # wait for errors set by reraise method
            time.sleep(1)
            if self.exception:
                # pylint: disable=raising-bad-type
                raise self.exception
            raise ConnectionError(f"{self.room} is not connected")

        thread = get_thread_ident()
        with self.lock:
            listener = self.listeners[thread]
        listener.add(event_type, callback)
        # use "initial_files" event to listen for whole filelist on room join
        self.process_queues()

    def enqueue_data(self, event_type, data):
        """Enqueue a data item for specific event type"""

        with self.lock:
            listeners = self.listeners.values()
            for listener in listeners:
                listener.enqueue(event_type, data)
                self.must_process = True

    @property
    def queues_enabled(self):
        """Whether queue processing is enabled"""

        return self.__queues_enabled

    @queues_enabled.setter
    def queues_enabled(self, value):
        """Sets whether queue processing is enabled"""

        with self.lock:
            self.__queues_enabled = value

    def process_queues(self, forced=False):
        """Process queues if any have data queued"""

        with self.lock:
            if (not forced and not self.must_process) or not self.queues_enabled:
                return
            self.must_process = False
        ARBITRATOR.awaken()

    @property
    def __listeners_for_thread(self):
        """All Listeners for the current thread"""

        thread = get_thread_ident()
        with self.lock:
            return [l for tid, l in self.listeners.items() if tid == thread]

    def validate_listeners(self):
        """Validates that some listeners are actually registered"""

        if self.exception:
            # pylint: disable=raising-bad-type
            raise self.exception

        listeners = self.__listeners_for_thread
        if not sum(len(l) for l in listeners):
            raise ValueError("No active listeners")

    def listen(self):
        """Listen for changes in all registered listeners."""

        self.validate_listeners()
        with ARBITRATOR.condition:
            while self.connected:
                ARBITRATOR.condition.wait()
                if not self.run_queues():
                    break

    def run_queues(self):
        """Run all queues that have data queued"""

        if self.exception:
            # pylint: disable=raising-bad-type
            raise self.exception
        listeners = self.__listeners_for_thread
        return sum(l.process() for l in listeners) > 0


class Room:
    """ Use this to interact with a room as a user
    Example:
        with Room("BEEPi", "SameFag") as r:
            r.post_chat("Hello, world!")
            r.upload_file("onii-chan.ogg")
    """

    # pylint: disable=unused-argument

    def __init__(
        self, name=None, user=None, password=None, key=None, subscribe=True, other=None
    ):
        """name is the room name, if none then makes a new room
        user is your user name, if none then generates one for you"""

        self.name = name
        self.password = password or ""
        self.key = key or ""
        self.admin = self.staff = self.owner = self.janitor = False
        self.__user_info = dict()
        self.__user_count = 0
        self.__files = OrderedDict()
        self.__upload_count = 0
        self.__room_score = 0.0

        self.config = Config()
        self.conn = Connection(self)
        if other:
            self.conn.cookies.update(other.conn.cookies)
        try:
            self.room_id, owner, checksum = self.__get_config()

            if not subscribe and not user:
                if other and other.user and other.user.name:
                    user = other.user.name
                else:
                    user = random_id(6)
            self.user = User(user, self.conn, self.config.max_nick)
            self.conn.connect(
                username=user, checksum=checksum, password=self.password, key=self.key
            )
            if not owner or self.user.nick is None:
                self.owner = False
            else:
                # can't really be abused because you can only manage stuff as logged user
                self.owner = bool(owner.lower() == self.user.nick.lower())

            if subscribe and other and other.user and other.user.logged_in:
                self.user.login_transplant(other.user)

            # check for first exception ever
            if self.conn.exception:
                # pylint: disable=raising-bad-type
                raise self.conn.exception
        except Exception:
            self.close()
            raise

    def __add_prop(self, key, admin=False):
        """Add gettable and settable room config property during runtime"""

        def getter(self):
            return self.config[key]

        def setter(self, val):
            if admin and not self.admin:
                raise RuntimeError(
                    f"You can't set the {key} key without mod privileges"
                )
            self.__set_config_value(self.config.get_real_key(key), val)

        setattr(self.__class__, key, property(getter, setter))

    def __set_config_value(self, key, value):
        """Sets a value for a room config"""

        self.check_owner()
        params = {"room": self.room_id, "config": to_json({key: value})}
        resp = self.conn.make_api_call("setRoomConfig", params)
        if "error" in resp:
            raise RuntimeError(f"{resp['error'].get('message') or resp['error']}")
        return resp

    def __get_config(self):
        """ Really connect """

        if not self.name:
            room_resp = self.conn.get(BASE_URL + "/new")
            room_resp.raise_for_status()
            url = room_resp.url
            try:
                self.name = re.search(r"r/(.+?)$", url).group(1)
            except Exception:
                raise IOError("Failed to create room")
        params = {"room": self.name}
        if self.key:
            params["roomKey"] = self.key
        if self.password:
            params["password"] = self.password
        config = self.conn.make_api_call("getRoomConfig", params)
        if "error" in config:
            raise RuntimeError(
                f"Failed to get room config for {self.name}\n"
                f"{config['error'].get('message') or config['error']}"
            )
        self.config.update(config)
        self.__add_prop("private")
        self.__add_prop("title")
        self.__add_prop("motd")
        self.__add_prop("adult")
        self.__add_prop("disabled", True)
        self.__add_prop("file_ttl", True)
        return (self.config.room_id, self.config.owner, config["checksum2"])

    def __repr__(self):
        return f"<Room({self.name}, {self.user.nick}, connected={self.connected})>"

    def __enter__(self):
        return self

    def __exit__(self, _extype, _value, _traceback):
        self.close()

    def __del__(self):
        self.close()

    @property
    def connected(self):
        """Room is connected"""

        return bool(hasattr(self, "conn") and self.conn.connected)

    def add_listener(self, event_type, callback):
        """Add a listener for specific event type.
        You'll need to actually listen for changes using the listen method"""

        return self.conn.add_listener(event_type, callback)

    def listen(self, once=False):
        """Listen for changes in all registered listeners.
        Use add_listener before calling this funcion to listen for desired
        events or set `once` to True to listen for initial room information """

        if once:
            # we listen for time event and return false so our
            # run_queues function will be also falsy and break the loop
            self.add_listener("time", lambda _: False)
        return self.conn.listen()

    @property
    def room_score(self):
        """Returns your room score if you are its owner.
        Room score of 1 and higher grants you VolaPro™"""

        if not self.owner:
            raise RuntimeError("You must own this room in order to check your score!")
        return self.__room_score

    @property
    def user_count(self):
        """Returns number of users in this room"""

        return self.__user_count

    @user_count.setter
    def user_count(self, count):
        """Sets user count value"""

        self.__user_count = count

    def __expire_files(self):
        """Because files are always unclean"""

        self.__files = OrderedDict(
            item for item in self.__files.items() if not item[1].expired
        )

    @property
    def files(self):
        """Returns copied list of File objects for this room.
        Note: This will only reflect the files at the time
        this method was called."""

        self.__expire_files()
        return list(self.__files.values())

    @property
    def filedict(self):
        """Returns dict of File objects for this room.
        Note: This will only reflect the files at the time
        this method was called."""

        self.__expire_files()
        return self.__files

    @filedict.setter
    def filedict(self, kv):
        """Updates filedict with single file entry or deletes given
        key if the value is False. Shouldn't be used by the user."""

        k, v = kv
        if v is not None:
            self.__files.update({k: v})
        else:
            with suppress(KeyError):
                del self.__files[k]

    def get_user_stats(self, name):
        """Return data about the given user. Returns None if user
        does not exist."""

        req = self.conn.get(BASE_URL + "/user/" + name)
        if req.status_code != 200 or not name:
            return None

        return self.conn.make_api_call("getUserInfo", {"name": name})

    def post_chat(self, msg, is_me=False, is_a=False):
        """Posts a msg to this room's chat. Set me=True if you want to /me"""

        if len(msg) > self.config.max_message:
            raise ValueError(
                f"Chat message must be at most {self.config.max_message} characters."
            )
        while not self.user.nick:
            with ARBITRATOR.condition:
                ARBITRATOR.condition.wait()
        if is_a:
            if not self.admin and not self.staff:
                raise RuntimeError("Can't modchat if you're not a mod or trusted")
            self.conn.make_call("command", self.user.nick, "a", msg)
            return
        if is_me:
            self.conn.make_call("command", self.user.nick, "me", msg)
            return

        self.conn.make_call("chat", self.user.nick, msg)

    def upload_file(
        self,
        filename,
        upload_as=None,
        blocksize=None,
        callback=None,
        information_callback=None,
        allow_timeout=False,
    ):
        """Uploads a file with given filename to this room.
        You may specify upload_as to change the name it is uploaded as.
        You can also specify a blocksize and a callback if you wish.
        Returns the file's id on success and None on failure."""

        with delayed_close(
            filename if hasattr(filename, "read") else open(filename, "rb")
        ) as file:
            filename = upload_as or os.path.split(filename)[1]
            try:
                file.seek(0, 2)
                if file.tell() > self.config.max_file:
                    raise ValueError(
                        f"File must be at most {self.config.max_file >> 30} GB"
                    )
            finally:
                try:
                    file.seek(0)
                except Exception:
                    pass

            files = Data(
                {"file": {"name": filename, "value": file}},
                blocksize=blocksize,
                callback=callback,
            )

            headers = {"Origin": BASE_URL}
            headers.update(files.headers)

            while True:
                key, server, file_id = self._generate_upload_key(
                    allow_timeout=allow_timeout
                )
                info = dict(
                    key=key,
                    server=server,
                    file_id=file_id,
                    room=self.room_id,
                    filename=filename,
                    len=files.len,
                    resumecount=0,
                )
                if information_callback:
                    if information_callback(info) is False:
                        continue
                break

            params = {"room": self.room_id, "key": key, "filename": filename}

            if self.key:
                params["roomKey"] = self.key
            if self.password:
                params["password"] = self.password

            while True:
                try:
                    post = self.conn.post(
                        f"https://{server}/upload",
                        params=params,
                        data=files,
                        headers=headers,
                    )
                    post.raise_for_status()
                    break

                except requests.exceptions.ConnectionError as ex:
                    if "aborted" not in repr(ex):  # ye, that's nasty but "compatible"
                        raise
                    try:
                        resume = self.conn.get(
                            f"https://{server}/rest/uploadStatus",
                            params={"key": key, "c": 1},
                        ).text
                        resume = from_json(resume)
                        resume = resume["receivedBytes"]
                        if resume <= 0:
                            raise ConnectionError("Cannot resume")
                        file.seek(resume)
                        files = Data(
                            {"file": {"name": filename, "value": file}},
                            blocksize=blocksize,
                            callback=callback,
                            logical_offset=resume,
                        )
                        headers.update(files.headers)
                        params["startAt"] = resume
                        info["resumecount"] += 1
                        if information_callback:
                            information_callback(info)
                    except requests.exceptions.ConnectionError as iex:
                        # ye, that's nasty but "compatible"
                        if "aborted" not in repr(iex):
                            raise
                        continue  # another day, another try
            return file_id

    def close(self):
        """Close connection to this room"""

        if hasattr(self, "conn"):
            self.conn.close()
            del self.conn
        if hasattr(self, "user"):
            del self.user

    def report(self, reason=""):
        """Reports this room to moderators with optional reason."""

        self.conn.make_call("submitReport", {"reason": reason})

    @property
    def user_info(self):
        """Get info of a user that gets updated through userInfo calls"""

        return self.__user_info

    @user_info.setter
    def user_info(self, kv):
        """Sets user_info dict entry through a tuple."""

        key, value = kv
        self.__user_info[key] = value

    def check_owner(self):
        """Helper method that rises an error if we don't have enough
        privileges to do owner related tasks."""

        if (not self.owner) and (not self.admin) and (not self.janitor):
            raise RuntimeError("Not enough auths do do that, big boy")

    def check_admin(self):
        """Helper method that rises an error if we do tasks that
        require admin/mod privileges and we aren't mod ourselves."""

        if not self.admin:
            raise RuntimeError("You must be an admin to do that")

    def clear(self):
        """Clears the cached information, if any."""

        self.__files.clear()

    def fileinfo(self, fid):
        """Ask lain about what he knows about given file. If the given file
        exists in the file dict, it will get updated. If the file was updated
        already, this will return File instance of it."""

        if not isinstance(fid, str):
            raise TypeError("Your file ID must be a string")
        has_fid = fid in self.__files
        if has_fid and self.__files[fid].updated:
            return self.__files[fid]
        try:
            info = self.conn.make_call_with_cb("getFileinfo", fid).result(5)
            if not info:
                warnings.warn(
                    f"Your query for file with ID: '{fid}' failed.", RuntimeWarning
                )
            elif has_fid:
                self.__files[fid].fileupdate(info)
        except (FutureTimeoutError, asyncio.CancelledError) as ex:
            raise ValueError(
                "lain didn't produce a callback!\n"
                "Are you sure your query wasn't malformed?"
            ) from ex
        return info

    def _generate_upload_key(self, allow_timeout=False):
        """Generates a new upload key"""

        # Wait for server to set username if not set already.
        while not self.user.nick:
            with ARBITRATOR.condition:
                ARBITRATOR.condition.wait()
        while True:
            params = {
                "name": self.user.nick,
                "room": self.room_id,
                "c": self.__upload_count,
            }
            if self.key:
                params["roomKey"] = self.key
            if self.password:
                params["password"] = self.password
            info = self.conn.make_api_call("getUploadKey", params)
            self.__upload_count += 1
            try:
                return info["key"], info["server"], info["file_id"]
            except Exception:
                to = int(info.get("error", {}).get("info", {}).get("timeout", 0))
                if to <= 0 or not allow_timeout:
                    raise IOError(f"Failed to retrieve key {info}")
                time.sleep(to / 10000)

    def delete_files(self, ids):
        """Remove one or more files"""

        self.check_owner()
        if not isinstance(ids, list):
            raise TypeError("You must specify list of files to delete!")
        self.conn.make_call("deleteFiles", ids)

    def transfer_owner(self, new_owner):
        """You had good run at it, it's time for someone else to get dirty"""

        if not self.owner and not self.admin:
            raise RuntimeError("You need more street creed to do this")

        new_owner = new_owner.strip().lower()
        if not new_owner:
            raise ValueError("Empty strings cannot be new owners")

        self.__set_config_value("owner", new_owner)

    def add_janitor(self, janitor):
        """Add janitor to the room"""

        if not self.owner and not self.admin:
            raise RuntimeError("Not enough street creed to do this")

        janitor = janitor.strip().lower()
        if not janitor:
            raise ValueError("Empty strings cannot be janitors")

        if janitor in self.config.janitors:
            return

        self.config.janitors.append(janitor)
        self.__set_config_value("janitors", self.config.janitors)

    def remove_janitor(self, janitor):
        """Remove janitor from the room"""

        if not self.owner and not self.admin:
            raise RuntimeError("Not enough street creed to do this")

        janitor = janitor.strip().lower()
        if not janitor:
            raise ValueError("Empty strings cannot be janitors")

        if janitor not in self.config.janitors:
            return

        self.config.janitors.remove(janitor)
        self.__set_config_value("janitors", self.config.janitors)

    def ban(self, nick="", address="", hours=6, reason="spergout", options=None):
        self.check_admin()
        if nick == "" and address == "":
            return
        who = []
        options = options or {}
        if address != "":
            if isinstance(address, str):
                who.append({"ip": address})
            if isinstance(address, list):
                for a in address:
                    who.append({"ip": a})
        if nick != "":
            if isinstance(nick, str):
                who.append({"user": nick})
            if isinstance(nick, list):
                for n in nick:
                    who.append({"user": n})
        ropts = {
            "ban": False,
            "hellban": False,
            "mute": False,
            "purgeFiles": False,
            "hours": hours,
            "reason": reason,
        }
        ropts.update(options)
        self.conn.make_call("banUser", who, ropts)

    def unban(self, nick="", address="", reason="", options=None):
        self.check_admin()
        if nick == "" and address == "":
            return
        who = []
        options = options or {}
        if address != "":
            if isinstance(address, str) and address != "":
                who.append({"ip": address})
            if isinstance(address, list):
                for a in address:
                    who.append({"ip": a})
        if nick != "":
            if isinstance(nick, str):
                who.append({"user": nick})
            if isinstance(nick, list):
                for n in nick:
                    who.append({"user": n})
        ropts = {
            "ban": True,
            "hellban": True,
            "mute": True,
            "timeout": True,
            "reason": reason,
        }
        ropts.update(options)
        self.conn.make_call("unbanUser", who, ropts)


def listen_many(*rooms):
    """Listen for changes in all registered listeners in all specified rooms"""

    rooms = set(r.conn for r in rooms)
    for room in rooms:
        room.validate_listeners()
    with ARBITRATOR.condition:
        while any(r.connected for r in rooms):
            ARBITRATOR.condition.wait()
            rooms = [r for r in rooms if r.run_queues()]
            if not rooms:
                return



#!/usr/bin/env python3

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

import re

from volapi import Room


def listen(room):
    """Open a volafile room and start listening to it"""
    def onmessage(m):
        """Print the new message and respond to it."""
        print(m)
        if m.admin or m.nick == r.user.name:
            return
        if "parrot" in m.msg.lower():
            r.post_chat("ayy lmao")
        elif m.msg.lower() in ("lol", "lel", "kek"):
            r.post_chat("*kok")
        else:
            r.post_chat(re.sub(r"\blain\b", "purpleadmin", m.msg, re.I))

    with Room(room) as r:
        r.user.change_nick("DumbParrot")
        r.add_listener("chat", onmessage)
        r.listen()

if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print("Provide a room ID", file=sys.stderr)
        sys.exit(1)

    while True:
        try:
            listen(sys.argv[1])
        except Exception as ex:
            print("Error listening", ex, file=sys.stderr)
        except KeyboardInterrupt:
            print("\nUser canceled this boat!!!1!", file=sys.stderr)
            sys.exit(0)



import logging
import warnings

from functools import partial


from .file import File
from .chat import ChatMessage
from .auxo import ARBITRATOR

LOGGER = logging.getLogger(__name__)

GENERICS = [
    "roomScore",
    "submitChat",
    "submitCommand",
    "pro",
    "room_old",
    "upload",
    "reconnectTimeout",
]


class Handler:
    def __init__(self, conn):
        self.conn = conn
        self.room = conn.room
        self.__head = "_handle_"
        self.__callbacks = {}
        self.__cid = 0
        for g in GENERICS:
            setattr(self, f"{self.__head}{g}", partial(self._handle_generic, g))

    def add_data(self, rawdata):
        """Add data to given room's state"""

        for data in rawdata:
            try:
                item = data[0]
                if item[0] == 2:
                    # Flush messages but we got nothing to flush
                    continue
                if item[0] != 0:
                    warnings.warn(f"Unknown message type '{item[0]}'", Warning)
                    continue
                item = item[1]
                # convert target to string because error codes are ints
                target = str(item[0])
                try:
                    data = item[1]
                except IndexError:
                    data = {}
                try:
                    method = getattr(self, f"{self.__head}{target}")
                    method(data)
                except AttributeError:
                    self._handle_unhandled(target, data)
            except IndexError:
                LOGGER.warning("Wrongly constructed message received: %r", data)

        self.conn.process_queues()

    def register_callback(self):
        """Register callback that we will have to wait for"""

        cid = str(self.__cid)
        self.__cid += 1
        event = ARBITRATOR.loop.create_future()
        self.__callbacks[cid] = event
        return cid, event

    def _handle_generic(self, target, data):
        """Handle generic notifications"""

        self.conn.enqueue_data(target, data)

    @staticmethod
    def _handle_unhandled(target, data):
        """Handle life, the universe and the rest"""

        warnings.warn(f"unknown data type '{target}' with data '{data}'", Warning)

    def _handle_401(self, data):
        """Handle Lain being helpful"""

        ex = ConnectionError(
            "Can't login to a protected room without a proper password", data
        )
        self.conn.reraise(ex)

    def _handle_429(self, data):
        """Handle Lain being helpful"""

        ex = IOError("Too fast", data)
        self.conn.reraise(ex)

    def _handle_callback(self, data):
        """Handle lain's callback. Only used with getFileinfo so far"""

        cb_id = data.get("id")
        args = data.get("args")
        event = self.__callbacks.pop(cb_id, None)
        if not event:
            return
        if not args:
            event.cancel()
            return
        err, info = args
        if err is None:
            event.set_result(info)
        else:
            LOGGER.warning("Callback returned error of %s", str(err))
            event.set_result(err)

    def _handle_userCount(self, data):
        """Handle user count changes"""

        self.room.user_count = data
        self.conn.enqueue_data("user_count", self.room.user_count)

    def _handle_userInfo(self, data):
        """Handle user information"""

        for k, v in data.items():
            if k == "nick":
                if v == "None":
                    v = "Volaphile"
                setattr(self.room.user, k, v)
                self.conn.enqueue_data(k, self.room.user.nick)
            elif k != "profile":
                if not hasattr(self.room, k):
                    warnings.warn(f"Skipping unset property {k}", ResourceWarning)
                    continue
                setattr(self.room, k, v)
                self.conn.enqueue_data(k, getattr(self.room, k))
            self.room.user_info = k, v
        self.conn.enqueue_data("user_info", self.room.user_info)

    def _handle_removeMessages(self, data):
        """Handle mods purging stuff"""

        try:
            removed_msgs = data.get("msgIds")
        except AttributeError:
            removed_msgs = []
        self.conn.enqueue_data("removed_messages", removed_msgs)

    def _handle_key(self, data):
        """Handle keys"""

        self.room.key = data
        self.conn.enqueue_data("key", self.room.key)

    def _handle_config(self, data):
        """Handle initial config push and config changes"""

        self.room.config.update(data)
        self.conn.enqueue_data("config", data)

    def _handle_files(self, data):
        """Handle new files being uploaded"""

        initial = data.get("set", False)
        files = data["files"]
        for f in files:
            try:
                fobj = File(
                    self.room,
                    self.conn,
                    f[0],
                    f[1],
                    type=f[2],
                    size=f[3],
                    expire_time=int(f[4]) / 1000,
                    uploader=f[6].get("nick") or f[6].get("user"),
                )
                self.room.filedict = fobj.fid, fobj
                if not initial:
                    self.conn.enqueue_data("file", fobj)
            except Exception:
                import pprint

                LOGGER.exception("bad file")
                pprint.pprint(f)
        if initial:
            self.conn.enqueue_data("initial_files", self.room.filedict.values())

    def _handle_delete_file(self, data):
        """Handle files being removed"""

        file = self.room.filedict.get(data)
        if file:
            self.room.filedict = data, None
            self.conn.enqueue_data("delete_file", file)

    def _handle_chat(self, data):
        """Handle chat messages"""

        self.conn.enqueue_data(
            "chat", ChatMessage.from_data(self.room, self.conn, data)
        )

    def _handle_changed_config(self, change):
        """Handle configuration changes"""

        key, value = change.get("key"), change.get("value")
        self.room.config.update({key: value})
        self.conn.enqueue_data("config", self.room.config)

    def _handle_chat_name(self, data):
        """Handle user name changes"""

        self.room.user.nick = data
        self.conn.enqueue_data("user", self.room.user)

    def _handle_time(self, data):
        """Handle time changes"""

        self.conn.enqueue_data("time", data / 1000)

    def _handle_subscribed(self, data):
        """Room is armed and ready to get DDoSed"""

        LOGGER.debug("armed and ready to receive messages")
        self.conn.enqueue_data("subscribed", data)



from enum import Enum

from .utils import html_to_text


class Roles(Enum):
    """All recognized roles"""

    WHITE = "white"
    USER = "green"
    PRO = "pro"
    OWNER = "owner"
    JANITOR = "janitor"
    DONOR = "donor"
    STAFF = "trusted user"
    ADMIN = "admin"
    SYSTEM = "system"

    @classmethod
    def from_options(cls, options):
        role_set = set()
        user = "profile" in options
        if user:
            if "admin" in options:
                role_set.add(cls.ADMIN)
            if "staff" in options:
                role_set.add(cls.STAFF)
            if "owner" in options:
                role_set.add(cls.OWNER)
            if "janitor" in options:
                role_set.add(cls.JANITOR)
            if "pro" in options:
                role_set.add(cls.PRO)
            if "donator" in options:
                role_set.add(cls.DONOR)
            if "user" in options:
                role_set.add(cls.USER)
        else:
            if "admin" in options or "staff" in options:
                role_set.add(cls.SYSTEM)
        if not role_set:
            role_set.add(cls.WHITE)
        return role_set

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class ChatMessage(str):
    """Basically a struct for a chat message. self holds the
    text of the message, files is a list of Files that were
    linked in the message, and rooms are a list of room
    linked in the message. There are also flags for whether the
    user of the message was logged in, a donor, or an admin."""

    # pylint: disable=no-member

    def __new__(cls, room, conn, nick, msg, roles=None, options=None, **kw):
        obj = super().__new__(cls, msg)
        obj.room = room
        obj.conn = conn
        obj.nick = nick
        obj.roles = roles or {Roles.WHITE}
        for entry in obj.roles:
            if entry not in Roles:
                raise ValueError("Invalid role")
        obj.options = options or {}

        # Optionals
        obj.files = kw.get("files", [])
        obj.rooms = kw.get("rooms", {})
        obj.data = kw.get("data", {})
        obj.mymsg = obj.data.get("self", False)
        return obj

    @staticmethod
    def from_data(room, conn, data):
        """Construct a ChatMessage instance from raw protocol data"""
        files = []
        rooms = {}
        msg = ""

        for part in data["message"]:
            ptype = part["type"]
            if ptype == "text":
                val = part["value"]
                msg += val
            elif ptype == "break":
                msg += "\n"
            elif ptype == "file":
                fileid = part["id"]
                fileobj = room.filedict.get(fileid)
                if fileobj:
                    files += (fileobj,)
                fileid = f"@{fileid}"
                msg += fileid
            elif ptype == "room":
                roomid = part["id"]
                rooms[roomid] = part["name"]
                msg += f"#{roomid}"
            elif ptype == "url":
                msg += part["text"]
            elif ptype == "raw":
                msg += html_to_text(part["value"])
            else:
                import warnings

                warnings.warn(f"unknown message type '{ptype}'", Warning)

        nick = data.get("nick") or data.get("user")
        options = data.get("options", {})
        data = data.get("data", {})

        message = ChatMessage(
            room,
            conn,
            nick,
            msg,
            roles=Roles.from_options(options),
            options=options,
            data=data,
            files=files,
            rooms=rooms,
        )
        return message

    def timeout(self, duration=3600):
        self.room.check_owner()
        if duration <= 0:
            raise RuntimeError("Timeout duration have to be a positive number")
        msg_id = self.id
        if msg_id is None:
            raise RuntimeError("No message ID, you can't timeout system or mods")
        self.conn.make_call("timeoutChat", msg_id, duration)

    @property
    def white(self):
        return Roles.WHITE in self.roles

    @property
    def user(self):
        return Roles.USER in self.roles

    @property
    def pro(self):
        return Roles.PRO in self.roles

    @property
    def owner(self):
        return Roles.OWNER in self.roles

    @property
    def janitor(self):
        return Roles.JANITOR in self.roles

    @property
    def donor(self):
        return Roles.DONOR in self.roles

    @property
    def green(self):
        return self.user and not self.purple

    @property
    def staff(self):
        return Roles.STAFF in self.roles

    @property
    def admin(self):
        return Roles.ADMIN in self.roles

    @property
    def purple(self):
        return self.admin or self.staff

    @property
    def system(self):
        return Roles.SYSTEM in self.roles

    @property
    def logged_in(self):
        return self.green or self.purple

    @property
    def ip_address(self):
        """Returns the uploader ip if available"""

        return self.data.get("ip")

    @property
    def id(self):
        """Returns id of the given chat message. Usefull if we want
        to check which messages are being removed by mods."""

        return self.data.get("id")

    def __repr__(self):
        prefix = ""
        if self.purple:
            prefix += "@"
        if self.owner:
            prefix += "👑"
        if self.pro:
            prefix += "✡"
        if self.janitor:
            prefix += "🧹"
        if self.donor:
            prefix += "💰"
        if self.green:
            prefix += "+"
        if self.system:
            prefix += "%"
        return f"<Msg({prefix}{self.nick}, {self})>"



"""
The MIT License (MIT)
Copyright © 2015 RealDolos
See LICENSE
"""

import random
import string

from contextlib import contextmanager
from html.parser import HTMLParser

try:
    import orjson as json
except ImportError:
    import json


class MLStripper(HTMLParser):
    """Used for stripping HTML from text."""

    def __init__(self):
        super().__init__()
        self.fed = []

    def handle_data(self, data):
        self.fed.append(data)

    def get_data(self):
        """Gets the non-HTML data from text that was fed in"""

        return "".join(self.fed)


def html_to_text(html):
    """Strips HTML tags from given text and returns it."""

    stripper = MLStripper()
    stripper.feed(html)
    return stripper.get_data()


def random_id(length):
    """Generates a random ID of given length"""

    def char():
        """Generate single random char"""

        return random.choice(string.ascii_letters + string.digits)

    return "".join(char() for _ in range(length))


def to_json(obj):
    """Create a compact JSON string from an object"""

    byte_dump = json.dumps(obj)
    if isinstance(byte_dump, bytes):
        return byte_dump
    return byte_dump.encode("utf8")


def from_json(string):
    """Create a Python object from a JSON string"""

    return json.loads(string)


@contextmanager
def delayed_close(closable):
    """Delay close until this contextmanager dies"""

    close = getattr(closable, "close", None)
    if close:
        # we do not want the library to close file in case we need to
        # resume, hence make close a no-op
        # pylint: disable=unused-argument
        def replacement_close(*args, **kw):
            """ No op """

        # pylint: enable=unused-argument
        setattr(closable, "close", replacement_close)
    try:
        yield closable
    finally:
        if close:
            setattr(closable, "close", close)
            closable.close()



"""
The MIT License (MIT)
Copyright © 2015 RealDolos
Copyright © 2018 Szero
See LICENSE
"""
# pylint: disable=bad-continuation,broad-except

import logging
import sys
import asyncio

from collections import namedtuple, defaultdict
from functools import wraps
from threading import Thread, Event, RLock, Condition, Barrier, get_ident
from urllib.parse import urlsplit
from copy import copy

from requests import Request
from requests.cookies import get_cookie_header
from autobahn.asyncio.websocket import WebSocketClientFactory
from autobahn.asyncio.websocket import WebSocketClientProtocol


logger = logging.getLogger(__name__)


def call_async(func):
    """Decorates a function to be called async on the loop thread"""

    @wraps(func)
    def wrapper(self, *args, **kw):
        """Wraps instance method to be called on loop thread"""

        def call():
            """Calls function on loop thread"""
            try:
                func(self, *args, **kw)
            except Exception:
                logger.exception(
                    "failed to call async [%r] with [%r] [%r]", func, args, kw
                )

        self.loop.call_soon_threadsafe(call)

    return wrapper


def call_sync(func):
    """Decorates a function to be called sync on the loop thread"""

    @wraps(func)
    def wrapper(self, *args, **kw):
        """Wraps instance method to be called on loop thread"""

        # Just return when already on the event thread
        if self.thread.ident == get_ident():
            return func(self, *args, **kw)

        barrier = Barrier(2)
        result = None
        ex = None

        def call():
            """Calls function on loop thread"""
            nonlocal result, ex
            try:
                result = func(self, *args, **kw)
            except Exception as exc:
                ex = exc
            finally:
                barrier.wait()

        self.loop.call_soon_threadsafe(call)
        barrier.wait()
        if ex:
            raise ex or Exception("Unknown error")
        return result

    return wrapper


class Awakener:
    """
    Callable helper thread to awaken non-event loop threads.
    The issue is that notify_all will temporarily release the
    lock to fire the listeners.

    If a listener would interact with volapi from there, this
    might mean a deadlock.
    Having the release on another thread will not block the
    event loop thread, therefore problem solved
    """

    def __init__(self, condition):
        self.condition = condition
        self.count = 0
        self.lock = RLock()
        self.event = Event()
        self.thread = Thread(daemon=True, target=self.target)
        self.thread.start()

    def __call__(self):
        with self.lock:
            self.count += 1
        self.event.set()

    def target(self):
        """Thread routine"""
        while self.event.wait():
            self.event.clear()
            while True:
                with self.lock:
                    if not self.count:
                        break
                    self.count -= 1
                with self.condition:
                    self.condition.notify_all()


class ListenerArbitrator:
    """Manages the asyncio loop and thread"""

    def __init__(self):
        self.loop = None
        self.condition = Condition()
        barrier = Barrier(2)
        self.awaken = Awakener(self.condition)
        self.thread = Thread(daemon=True, target=lambda: self._loop(barrier))
        self.thread.start()
        barrier.wait()

    def _loop(self, barrier):
        """Actual thread"""

        if sys.platform != "win32":
            self.loop = asyncio.new_event_loop()
        else:
            self.loop = asyncio.ProactorEventLoop()
        asyncio.set_event_loop(self.loop)
        barrier.wait()
        try:
            self.loop.run_forever()
        except Exception:
            sys.exit(1)

    @call_async
    def create_connection(self, room, ws_url, agent, cookies):
        """Creates a new connection"""

        urlparts = urlsplit(ws_url)
        req = Request("GET", ws_url)
        cookies = get_cookie_header(cookies, req)
        if cookies:
            headers = dict(Cookie=cookies)
        else:
            headers = None

        factory = WebSocketClientFactory(ws_url, headers=headers, loop=self.loop)
        factory.useragent = agent
        factory.protocol = lambda: room
        conn = self.loop.create_connection(
            factory,
            host=urlparts.netloc,
            port=urlparts.port or 443,
            ssl=urlparts.scheme == "wss",
        )
        asyncio.ensure_future(conn, loop=self.loop)

    def __send_message(self, proto, payload):
        # pylint: disable=no-self-use
        """Sends a message"""

        try:
            if not isinstance(payload, bytes):
                payload = payload.encode("utf-8")
            if not proto.connected:
                raise IOError("not connected")
            proto.sendMessage(payload)
            logger.debug("sent: %r", payload)
        except Exception as ex:
            logger.exception("Failed to send message")
            proto.reraise(ex)

    @call_sync
    def send_sync_message(self, proto, payload):
        self.__send_message(proto, payload)

    @call_async
    def send_async_message(self, proto, payload):
        self.__send_message(proto, payload)

    @call_sync
    def close(self, proto):
        # pylint: disable=no-self-use
        """Closes a connection"""

        try:
            proto.sendClose()
        except Exception as ex:
            logger.exception("Failed to send close")
            proto.reraise(ex)


ARBITRATOR = ListenerArbitrator()


class Listeners(namedtuple("Listeners", ("callbacks", "queue", "enlock", "lock"))):
    """Collection of Listeners
    `callbacks` are function objects.
    `queue` holds data that will be sent to each function object
    in `callbacks` variable.
    Each issue of `process` method will run given callback
    against items in `queue` when their types match. After that
    `queue` is cleared and amount of `callbacks` is returned.
    Callbacks that return False will be removed."""

    def __new__(cls):
        return super().__new__(
            cls, defaultdict(list), defaultdict(list), RLock(), RLock()
        )

    def process(self):
        """Process queue for these listeners. Only the items with type that
        matches """

        with self.lock, self.enlock:
            queue = copy(self.queue)
            self.queue.clear()
            callbacks = copy(self.callbacks)

        with self.lock:
            rm_cb = False
            for ki, vi in queue.items():
                if ki in self.callbacks:
                    for item in vi:
                        for cb in self.callbacks[ki]:
                            if cb(item) is False:
                                callbacks[ki].remove(cb)
                                if not callbacks[ki]:
                                    del callbacks[ki]
                                rm_cb = True

        with self.lock:
            if rm_cb:
                self.callbacks.clear()
                for k, v in callbacks.items():
                    self.callbacks[k].extend(v)
            return len(self.callbacks)

    def add(self, callback_type, callback):
        """Add a new listener"""

        with self.lock:
            self.callbacks[callback_type].append(callback)

    def enqueue(self, item_type, item):
        """Queue a new data item, make item iterable"""

        with self.enlock:
            self.queue[item_type].append(item)

    def __len__(self):
        """Return number of listeners in collection"""

        with self.lock:
            return len(self.callbacks)


class Protocol(WebSocketClientProtocol):
    """Implements the websocket protocol"""

    def __init__(self, conn):
        super().__init__()
        self.conn = conn
        self.connected = False
        self.max_id = 0
        self.send_count = 1
        self.session = None

    def onConnect(self, _response):
        self.connected = True

    async def onOpen(self):
        await self.conn.on_open()

    def onMessage(self, payload, isBinary):
        if not payload:
            logger.warning("empty frame!")
            return
        try:
            if not isBinary:
                payload = payload.decode("utf-8")
            self.conn.on_message(payload)
        except Exception:
            logger.exception("something went horribly wrong")

    async def onClose(self, _wasClean, _code, _reason):
        await self.conn.on_close()
        self.connected = False

    def connection_lost(self, exc):
        self.reraise(exc)
        self.connected = False

    def reraise(self, ex):
        if hasattr(self, "conn"):
            self.conn.reraise(ex)
        else:
            logger.error("Cannot reraise")

    def __repr__(self):
        return f"<Protocol({self.session},max_id={self.max_id},send_count={self.send_count})>"



"""
This file is part of Volapi.

Volapi is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Volapi is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Volapi.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
from setuptools import setup


def find_version(filename):
    """
    Search for assignment of __version__ string in given file and
    return what it is assigned to.
    """
    with open(filename, "r") as filep:
        version_file = filep.read()
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
        )
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")

setup(
    name="volapi",
    version=find_version("volapi/constants.py"),
    description="API for Volafile.org",
    long_description=open("README.rst", "r").read(),
    url="https://github.com/volafiled/python-volapi",
    license="MIT",
    author="RealDolos, szero",
    author_email="dolos@cock.li, singleton@tfwno.gf",
    packages=["volapi"],
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=[l.strip() for l in open("requirements.txt").readlines()],
)
